{*****************************************************************************
* UNIT DESCRIPTION:  Form class for charting
* SPEC REF: (if used)
* DESIGN REF: (if used)
*
* Copyright (c) 2009 Schweitzer Engineering Laboratories, Pullman, Washington
******************************************************************************}
unit FormChartTimeSeries;

interface

uses
  Windows, Messages, SysUtils, StrUtils, DateUtils,
  Controls, Tabs, Forms, Dialogs, Buttons, ExtCtrls, Graphics,
  Classes, Contnrs, ToolWin, ComCtrls, StdCtrls, ClipBrd,
  TeeEdit, TeEngine, TeeTools, TeeProcs, TeePolar, Chart, Series,
  ClassTimeSeries, ClassTimeSeriesChart,
  ClassTimeSeriesChartSSI, ClassTimeSeriesChartHRM,
  ClassChartInput,
  ClassChartInputSSI,  ClassChartInputHRM,
  TimeRoutines, Types,
  TimeSeriesConstants, TeeScroB; //, ClassLoadProfileFile_010;
const
  INVALID_CHART_MSG = 'Invalid chart type.';
  CRLF = Chr(13) + Chr(10);     // carriage return line feed
  CR = Chr(13);     // carriage return
  TAB = Chr(9);
  INTERVAL_COLOR = clWebLightSlateGray;
  PARENTEVENT_COLOR = clYellow;
  SUBEVENT_COLOR = clRed;
  EVENTMARKER_COLOR = clYellow;
  DATETIME_WITH_SEC = 'MM/dd/yy hh:mm:ss';
  TIME_WITH_MS = 'hh:mm:ss.zzz';
  TIME_WITH_SEC = 'hh:mm:ss';
  SHGFP_TYPE_CURRENT         = 0;
  CSIDL_PERSONAL = $0005;  // Doc and settings > User > My Documents
  SPECTRAL_LIMIT = 3300;
  HARMONIC_LIMIT = 50;
  MARKERS_PER_EVENT = 4;
  MA_ARRAY_SIZE = 8;  // maximum array size for mag/angle evaluation
type
  TChartMenuEvent = procedure (form : TForm) of object;  // method pointer
  TArrayOfArrayOfDouble = array of array of double;
  TArrayOfDouble = array of double;
  TArrayOfInteger = array of integer;
  TArrayOfSingle = array of single;
  TArrayOfColor = array of TColor;

  TfrmChartTimeSeries = class(TForm)
    PanelChart: TPanel;
    SplitterEventForm: TSplitter;
    ScrollBoxChartControls: TScrollBox;
    ToolBar1: TToolBar;
    SpeedButtonPointer: TSpeedButton;
    SpeedButtonRotate: TSpeedButton;
    SpeedButtonTranslate: TSpeedButton;
    SpeedButtonZoom: TSpeedButton;
    SpeedButtonThickness: TSpeedButton;
    SpeedButton3D: TSpeedButton;
    SpeedButtonEdit: TSpeedButton;
    SpeedButtonCopy: TSpeedButton;
    SpeedButtonReset: TSpeedButton;
    SpeedButtonCursor: TSpeedButton;
    SpeedButtonDateTime: TSpeedButton;
    ChartHMIDisplay: TChart;
    ChartTool1: TCursorTool;
    ChartHMICBEMA: TChart;
    ChartTool2: TCursorTool;
    SpeedButtonLabel: TSpeedButton;
    SpeedButtonInformation: TSpeedButton;
    ComboBoxDateFormat: TComboBox;
    ButtonImport: TButton;
    SpeedButtonPrint: TSpeedButton;
    ChartScrollBarXAxis: TChartScrollBar;
    procedure ButtonImportClick(Sender: TObject);
    procedure ChartHMIDisplayGetNextAxisLabel(Sender: TChartAxis;
      LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
    procedure ChartHMIDisplayMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
  private
    FChartType              : TTimeSeriesChartType;   // chart type: ctLDP, ctSSI
    FChartSubType           : TTimeSeriesChartSubType;  // Sub-type of chart
    FChartDirty             : boolean;   // true if chart type changed since last display
    FUpdateEnabled          : boolean;  // true if display may be updated
    FDeviceType             : TTimeSeriesDeviceType;  // device type: dtSEL7XX...
    FChartTimeSeries        : TTimeSeriesChart;
//    FDeviceInfoDialog       : TFormDeviceInfo;     // information window
    ChartSeries             : array of TCustomSeries;
    BarChartSeries          : array of TCustomBarSeries;
    PolarChartSeries        : array of TCustomPolarSeries;
    BarSeries               : TBarSeries;        // used for Harmonics
    PointSeries             : TPointSeries;  // used for LDP events
//    CBEMAFunctionSeries   : array [0..1] of TLineSeries;  // USE LATER
    FChartInputSettingsSSI  : TTimeSeriesInputSettingsSSI;
    FChartInputSettingsHRM  : TTimeSeriesInputSettingsHRM;
    FDateFormat             : string;  // date format string
    FMenuChart              : boolean;   // true if launched from tools menu
    FChartInstance          : integer;  // index identifying chart
    FChartAxisScale         : boolean;  // true if full harmonic axis, otherwise false
    FDefaultPath            : string;  // remember browse location during session
    FFileCount              : integer;
  public
    ChartMenuHandler : TChartMenuEvent; // delegate (method pointer)

    Constructor Create(aOwner: TComponent); reintroduce; overload;
    procedure SetChartType(aChartType : TTimeSeriesChartType);
    function  ReadChartData(aChartType : TTimeSeriesChartType;
                            fileReset, inputReset: boolean; filePath : string;
                            var ChartReset, ChartValid : boolean ) : boolean;
    procedure UpdateChart(reset : boolean);   overload;
    function  DisplayChartFile(filePath : string) : boolean;
    function  DisplayChart(aChartType : TTimeSeriesChartType; fileReset,inputReset: boolean; filePath : string; showChart : boolean ) : boolean; overload;
    procedure DisplayChart(found,chartReset,showChart,fileReset,inputReset,doCBEMA : boolean);  overload;
    procedure Display(reset : boolean);
    procedure InitChartMinMax(reset : boolean; var xMin,xMax,yMin,yMax : double);
    procedure UpdateChartMinMax(reset : boolean; xMin,xMax,yMin,yMax : double);
    procedure InitChartAttributes(channelCount : integer);
    procedure ChartHarmonicMagAnglePolar(index,channelCount : integer;
                myRecord : double;
                dataRecord : TTimeSeriesRecord);
    procedure ChartHarmonicMagAnglePolarLabels(channelCount : integer);
    function SkipHarmonicRecord(myRecord : double) : boolean;
    function GetCBEMA : boolean;
    procedure SetCBEMA(doCBEMA : boolean);
    procedure SetChartLineStyles();
    procedure SetDateTimeLabels();
    procedure ResetChartWindow;
    procedure ResetChartControls();
    function CBEMAChartVisible(DoCBEMA : boolean) : boolean;

    procedure DisplayMouseQueryFeedback;

    procedure FreeCharts();
    procedure AllocateChartArrays(
              channelCount : integer;
              harmSettings : TTimeSeriesInputSettingsHRM;
              var chartRec, eventRec, chartMarker, eventNumber : TArrayOfInteger;
              var eventColor : TArrayOfColor;
              var chartHeight, workArray : TArrayOfDouble;
              var xArray : TArrayOfDouble;
              var yArray : TArrayOfArrayOfDouble;
              var vbase, binsize : double;
              var doStatic : boolean);
    procedure FillChartArrays(channelCount, recordCount : integer;
              xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);
    procedure ChartLDPHarmonic(eventCount, channelCount : integer;
                doStatic : boolean; vbase, binSize : double;
                harmSettings : TTimeSeriesInputSettingsHRM;
                chartMarker, chartRec: TArrayOfInteger;
                chartHeight, workArray : TArrayOfDouble;
                xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);

    property ChartType : TTimeSeriesChartType read FChartType;
    property ChartSubType : TTimeSeriesChartSubType read FChartSubType write FChartSubType;
    property ChartDirty : boolean read FChartDirty;
    property UpdateEnabled : boolean read FUpdateEnabled;
    property DeviceType : TTimeSeriesDeviceType read FDeviceType;
    property ChartTimeSeries : TTimeSeriesChart read FChartTimeSeries;
//    property DeviceInfoDialog : TFormDeviceInfo read FDeviceInfoDialog write FDeviceInfoDialog;
    property ChartInputSettingsSSI : TTimeSeriesInputSettingsSSI read FChartInputSettingsSSI;
    property DateFormat : string read FDateFormat write FDateFormat;
    property MenuChart : boolean read FMenuChart write FMenuChart;
    property ChartInstance : integer read FChartInstance write FChartInstance;

  end;

implementation

Uses
  EditChar,
  TeeEditCha,
  Math;

{$R *.dfm}
{************************************************************************
* NAME:   Create
* FUNCTIONAL DESCRIPTION: class constructor
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
constructor TfrmChartTimeSeries.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FChartTimeSeries := TTimeSeriesChartHRM.Create(nil);
  FChartType := ctInvalid;    // nothing charted yet
  FChartSubType := cstInvalid;
  FChartDirty := true;
  FUpdateEnabled := true;
//  FDeviceInfoDialog := TFormDeviceInfo.Create(nil);
//  FDeviceInfoDialog.Parent :=  self;
//  FDeviceInfoDialog.Visible := false;
  FDateFormat := 'mm/dd/yyyy';

  self.HelpKeyword := 'Chart';

  if(ChartTool1 <> nil) then
  begin
    ChartTool1.Active := false;
    ChartTool1.Snap := true;
  end;
  if(ChartTool2 <> nil) then
  begin
    ChartTool2.Active := false;
    ChartTool2.Snap := true;
  end;
  FChartInstance := 0;
  FMenuChart := false;
  FChartAxisScale := false;
  FDefaultPath := '';
  FFileCount := 0;
end;

{************************************************************************
* NAME:     SetChartType
* FUNCTIONAL DESCRIPTION:   Set the chart type
* INPUT PARAMETERS:  aChartType : TTimeSeriesChartType  new chart type
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TfrmChartTimeSeries.SetChartType(aChartType : TTimeSeriesChartType);
begin
  {This function is used to change the chart type based on the incoming
  argument.  The argument indicates the chart type desired.  It is compared
  to the current chart type, and if different, the input settings for the
  current chart object will be saved and the charting data will be freed.
  A new chart object of the requested type will be created and the input
  settings corresponding to the requested chart type will be retrieved and
  made current.
  }

  if(FChartType <> aChartType) then
  begin
    case aChartType of
      ctUndefined,ctHarmonic:
        begin
          // If input values not saved previously, create a default set for LDP
          if(FChartInputSettingsHRM = nil) then
          begin
            FChartInputSettingsHRM := TTimeSeriesInputSettingsHRM.Create;  // preserved input settings
            FChartInputSettingsHRM.ChartAxisDateTime := false;
          end;
          // InputSettings is a container for the input of the active chart
          if(FChartTimeSeries.EventData.InputSettings = nil) then
          begin
            FChartTimeSeries.EventData.InputSettings := FChartInputSettingsHRM;
          end;

          FChartTimeSeries.SaveChartActiveChannels;  // save settings in previous chart
          FChartTimeSeries.SaveChartActiveEvents;    // save SSI Channels before switching

          FreeAndNil(FChartTimeSeries);
          FChartTimeSeries := TTimeSeriesChartHRM.Create(nil);

          FChartTimeSeries.ChartDisplay := ChartHMIDisplay;
          FChartTimeSeries.CursorTool := ChartTool1;
          FChartTimeSeries.EventData.DateFormat := DateFormat;
          // Make "InputSettings" a reference to the LDP input settings
          FChartTimeSeries.EventData.InputSettings := FChartInputSettingsHRM;
        end;

      ctSSI:
        begin
          // If input values not saved previously, create a default set for SSI
          if(FChartInputSettingsSSI = nil) then
          begin
            FChartInputSettingsSSI := TTimeSeriesInputSettingsSSI.Create;  // preserved input settings
          end;
          // InputSettings is a container for the input of the active chart
          if(FChartTimeSeries.EventData.InputSettings = nil) then
          begin
            FChartTimeSeries.EventData.InputSettings := FChartInputSettingsSSI;
          end;

          FChartTimeSeries.SaveChartActiveChannels;   // save settings in previous chart

          FreeAndNil(FChartTimeSeries);
          FChartTimeSeries := TTimeSeriesChartSSI.Create(nil);

          FChartTimeSeries.ChartDisplay := ChartHMIDisplay;
          FChartTimeSeries.CursorTool := ChartTool1;
          FChartTimeSeries.EventData.DateFormat := DateFormat;
          // Make "InputSettings" a reference to the SSI input settings
          FChartTimeSeries.EventData.InputSettings := FChartInputSettingsSSI;
        end;
      ctInvalid:
      begin
        // do nothing
      end
      else
      begin
        raise Exception.Create(INVALID_CHART_MSG);
      end;
    end;
    FChartType := aChartType;   // now update the chart type
  end;
end;

{************************************************************************
* NAME:     ResetChartWindow
* FUNCTIONAL DESCRIPTION:   Reset panels of chart window based on chart type
* INPUT PARAMETERS:  none
* OUTPUT PARAMETERS: none
* RETURN:            none
************************************************************************}
procedure TfrmChartTimeSeries.ResetChartWindow;
Const
  TABLIST : array [0..2] of string = ('General','Channels','Events');
  HARMONIC_LIST : string = '2,3,5,10,25';
  SPECTRAL_LIST : string = '10,50,60,100,120,150,180,200,240,500,600';
  MARGIN_SIZE = 3;
var
//  oldTab : integer;
  notHarmonic : boolean;
  isHarmonic : boolean;
  notPolar : boolean;
  harmonicInput : TTimeSeriesInputSettingsHRM;
begin
  notHarmonic := (FChartType <> ctHarmonic);
  isHarmonic := (FChartType = ctHarmonic);
  if(isHarmonic) then
  begin
    harmonicInput := FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsHRM;
    notPolar := not ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV));
  end else
  begin
    harmonicInput := nil;
    notPolar := true;  // display controls for all non-harmonic charts
  end;

  ComboBoxDateFormat.Visible := true;
//  ComboBoxDateFormat.Visible := notHarmonic;
  ChartScrollBarXAxis.Visible := isHarmonic;
  ChartScrollBarXAxis.Enabled := isHarmonic;
  SpeedButtonDateTime.Visible := notHarmonic;  // date/time labels not used for harmonic

  SpeedButtonRotate.Visible := notPolar;  // 3D rotation not used for polar
  SpeedButtonZoom.Visible := notPolar;  // 3D zoom not used for polar
  SpeedButtonTranslate.Visible := notPolar;  // 3D Translation not used for polar
  SpeedButtonThickness.Visible := notPolar;  // 3D thickness not used for polar
  SpeedButton3D.Visible := notPolar;  // 3D chart not used for polar

  if(FChartType = ctHarmonic) then
  begin
    ChartScrollBarXAxis.Visible := notPolar;
    ChartScrollBarXAxis.Enabled := notPolar;

    FChartTimeSeries.EventData.InputSettings.ChartLabels := false;
    FChartTimeSeries.ChartDisplay.BottomAxis.Grid.Visible:= false;

    SpeedButtonReset.Hint := 'Toggle chart x-axis limits.';
    harmonicInput.HarmonicChartDisplayStart := 0;
    harmonicInput.HarmonicChartDisplayInterval := 5;
  end else
  begin
    SpeedButtonReset.Hint := 'Reset chart to fill area.';
  end;

  if((FChartType = ctHarmonic) and (FChartSubType = cstHRM_INT)) then
  begin

    harmonicInput.HarmonicChartDisplayStart := 0;
    harmonicInput.HarmonicChartDisplayInterval := 60;
  end;

//  oldTab := FDeviceInfoDialog.TabControlDeviceInfo.TabIndex;  // original tab selection

  case FChartType of
    ctSSI:     // SSI always has chart and event summary
    begin
      SplitterEventForm.Align := alTop;
      PanelChart.Align := alClient;
//      FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Clear;
//      FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Add(TABLIST[0]);
//      FDeviceInfoDialog.TabControlDeviceInfo.TabIndex := 0;
    end;
    ctHarmonic:
    begin

//KWM      ChartDisplayChange;

//      FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Clear;
//      FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Add(TABLIST[0]);
//      FDeviceInfoDialog.TabControlDeviceInfo.TabIndex := 0;
      CBEMAChartVisible(false);
    end else  // unknown chart type: display only chart in entire window
    begin
//      FDeviceInfoDialog.TabControlDeviceInfo.Tabs.Clear;
      CBEMAChartVisible(false);
    end;
  end;
end;


function TfrmChartTimeSeries.ReadChartData(aChartType : TTimeSeriesChartType;
              fileReset, inputReset: boolean; filePath : string;
              var chartReset, chartValid : boolean ) : boolean;
const
  labelList : array[0..12] of string = ('Record','Date','Time','Ia','Ib','Ic','Ig','In','Va','Vb','Vc','Phase','Status');

var
  i,j,headerCount,dataCount : integer;
  dataRecord : TTimeSeriesRecord;
begin

  headerCount := 5;
  if(Pos('PIV',filePath) > 0) then
  begin
    SetChartType(ctHarmonic);
    dataCount := 7;
    ChartTimeSeries.EventData.ChartSubType := cstHRM_PIV;
  end else
  if(Pos('MAV',filePath) > 0) then
  begin
    SetChartType(ctHarmonic);
    dataCount := 9;
    ChartTimeSeries.EventData.ChartSubType := cstHRM_MAV;
  end else
  begin
    SetChartType(ctSSI);
    dataCount := 8;
    ChartTimeSeries.EventData.ChartSubType := cstInvalid;
  end;

  ChartTimeSeries.EventData.RawDataLabels.Clear();
  ChartTimeSeries.EventData.SetLengthHeaderColumns(headerCount);

  if(dataCount = 8) then
  begin
    ChartTimeSeries.EventData.HeaderColumns[0] := -1;
    ChartTimeSeries.EventData.HeaderColumns[1] := 1;
    ChartTimeSeries.EventData.HeaderColumns[2] := 2;
    ChartTimeSeries.EventData.HeaderColumns[3] := 11;
    ChartTimeSeries.EventData.HeaderColumns[4] := 12;
  end else
  begin
    for i := 0 to headerCount do
    begin
      ChartTimeSeries.EventData.HeaderColumns[i] := -1;
    end;
    ChartTimeSeries.EventData.HeaderColumns[0] := 0;
  end;

  ChartTimeSeries.EventData.SetLengthDataColumns(dataCount);
  for i := 0 to dataCount do
  begin
    ChartTimeSeries.EventData.DataColumns[i] := i+1;
    ChartTimeSeries.EventData.RawDataLabels.Add(labelList[i]);
  end;

  for i := 0 to 10  do
  begin
      dataRecord := TTimeSeriesRecord.Create;
      dataRecord.RecordNumber := i+1;

      if(dataCount = 8) then
      begin
        dataRecord.Phase := '...';
        dataRecord.Status := 'P';
        dataRecord.DateTime := Now;
        SetLength(dataRecord.Value,dataCount+1);
        dataRecord.Value[0] := i+2;
        for j := 1 to dataCount-1 do
        begin
          dataRecord.Value[j] := j;
        end;
      end else
      begin
        SetLength(dataRecord.Value,dataCount+1);
        dataRecord.Value[0] := i+2;
        for j := 1 to dataCount-1 do
        begin
          dataRecord.Value[j] := j;
        end;
      end;
      ChartTimeSeries.EventData.RawDataTable.Add(dataRecord);
  end;

  result := true;

end;

{!******************************************************************************
* MODULE NAME ...  ResetChartControls
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Reset chart controls based on input settings for current HMI context
*
* INPUT PARAMETERS:  None
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ResetChartControls();
begin
  SetDateTimeLabels;
end;

procedure TfrmChartTimeSeries.UpdateChart(reset : boolean);
begin
  FChartTimeSeries.SaveChartActiveChannels;   // remember which channels are active
  FChartTimeSeries.InitChartVariableCount;   // reset variable count if appropriate
  FChartTimeSeries.InitChartActiveChannelsDefault(false);   // re-allocate if necessary

  if(reset or ChartDirty) then   // chart type changed, reset charting controls
  begin
    ResetChartWindow;
  end;

    Display(reset);  // regenerate display based on currently checked events

end;

function TfrmChartTimeSeries.DisplayChartFile(filePath : string) : boolean;
var
  myChartType : TTimeSeriesChartType;
  myChartSubType : TTimeSeriesChartSubType;
  fileData : TStringList;
begin
  fileData := TStringList.Create;
  try
    if(Pos('PIV',filePath) > 0) then
    begin
      myChartType := ctHarmonic;
      myChartSubType := cstHRM_PIV;
    end else
    if(Pos('MAV',filePath) > 0) then
    begin
      myChartType := ctHarmonic;
      myChartSubType := cstHRM_MAV;
    end else
    begin
      myChartType := ctSSI;
      myChartSubType := cstInvalid;
    end;
    SetChartType(myChartType);
    FChartSubType := myChartSubType;
    ChartTimeSeries.EventData.ChartSubType := myChartSubType;
    result := DisplayChart(myChartType,true,true,filePath,true);
  finally
    fileData.Free;
  end;
end;

{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data found in file
* INPUT PARAMETERS:   aChartType : THMIChartType  type of chart to display
*                     fileReset : boolean  Initialize display of chart (true/false)
*                     inputReset : boolean  reset all user input parameters if true
*                     filename : string  is the original filename
*                     showChart : boolean  display the chart (true/false)
* OUTPUT PARAMETERS: none
* RETURN:             result : boolean  true if data could be parsed
************************************************************************}
function TfrmChartTimeSeries.DisplayChart(aChartType : TTimeSeriesChartType;
  fileReset,inputReset: boolean; filePath : string; showChart : boolean ) : boolean;
var
  chartReset, chartValid, found, doCBEMA : boolean;
  oldChartType : TTimeSeriesChartType;
  myInputSettings : TTimeSeriesInputSettingsSSI;
begin
  found := true;
  // If display is already in progress, just return
  if(FUpdateEnabled) then
  begin
    oldChartType := FChartType;  // original type
    chartReset := true;

    if(fileReset and inputReset) then    // cancel date/time filter controls
    begin
      if(FChartType = ctSSI) then
      begin
        myInputSettings := (ChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI);
        myInputSettings.BeginDateActive := false;
        myInputSettings.EndDateActive := false;
      end;
    end;

    doCBEMA := GetCBEMA;
    SetCBEMA(false);       // must disable CBEMA so that VSSI chart will be generated

    try

      // Read chart data from file  (chartReset true if fileReset or change chart type)
      found := ReadChartData(aChartType, fileReset, inputReset, filePath, chartReset, chartValid);
      if(found) then
      begin
        if(not chartValid) then
        begin
//          ShowSimpleMessageBox('Error reading data into chart.');
        end else
        begin
          FUpdateEnabled := false;  // don't update display until finished generating it
          try
            DisplayChart(found,chartReset,showChart,fileReset,inputReset,doCBEMA);
            if(aChartType <> oldChartType) then
            begin
              ResetChartWindow();
              FChartTimeSeries.AutoScaleAxes(true);
            end;

            FUpdateEnabled := true;   // It is safe to update now

            FChartTimeSeries.ChartDisplay.Update();

          finally
            if(found or (not chartReset and (FChartTimeSeries.EventData.RawDataTable.Count > 1))) then
            begin
            // successful chart
            end else
            begin      // something went wrong
              FChartType := oldChartType;  // restore original chart type
            end;
          end;
        end;
      end;
    finally
      FUpdateEnabled := true;   // It is safe to update now
      FChartType := oldChartType;  // restore original chart type
    end;
  end;

  result := found;
end;

{************************************************************************
* NAME:   DisplayChart
* FUNCTIONAL DESCRIPTION:  Display Chart data already loaded into memory
* INPUT PARAMETERS:   found : boolean   data is in memory
*                     chartReset : boolean true fileReset or ChartType changed
*                     fileReset : boolean  true if new file imported
*                     inputReset : boolean true if reset user input options
*                     showChart : boolean  display the chart (true/false)
*                     doCBEMA : boolean true if CBEMA is default chart
* OUTPUT PARAMETERS: none
* RETURN:             none
************************************************************************}
procedure TfrmChartTimeSeries.DisplayChart(found,chartReset,showChart,fileReset,inputReset,doCBEMA : boolean);
var
  oldState : TWindowState;
begin
  if(found) then      // valid charting data found in file
  begin
    UpdateChart(inputReset);   // update the chart
  end;

  if(found or (not chartReset and (FChartTimeSeries.EventData.RawDataTable.Count > 1))) then
  begin
    if(showChart or Visible) then           // Display the chart window
    begin
      // NOTE: Delphi has unexpected behavior when chart window has been maximized.
      // This can be avoided by displaying in normal state, then switching to
      // original state.
      oldState := WindowState;
      WindowState := wsNormal;   // cancel window maximization
      Show;
      if(oldState <> wsNormal) then  // restore to original window size
      begin
        WindowState := oldState;
      end;
      if(FChartType = ctHarmonic) then
      begin
        ChartHMIDisplay.SetFocus;
      end;
    end else
    begin
      SetCBEMA(doCBEMA);  // Set CBEMA flag in case it was disabled.
    end;
  end;
end;

{************************************************************************
* NAME:   Display
* FUNCTIONAL DESCRIPTION:  Create the chart, assign attributes and insert data for display
* INPUT PARAMETERS:
*                    reset: boolean  true if regenerating display (rescale axis limits),
*                                    false if updating display (retain current axis limits)
* OUTPUT PARAMETERS: none
* RETURN:             none
************************************************************************}
procedure TfrmChartTimeSeries.Display(reset : boolean);
var
  eventCount,channelCount : integer;
  xMin,xMax,yMin,yMax : double;
  harmSettings : TTimeSeriesInputSettingsHRM;
  chartMarker,chartRec,eventRec,eventNumber : TArrayOfInteger;   // event number of charted events
  chartHeight,workArray : TArrayOfDouble;
  eventColor : TArrayofColor;
  vbase,binSize : double;
  table : TArrayOfArrayOfDouble;
  xArray : TArrayOfDouble;
  doStatic : boolean;
begin
  if(ChartTimeSeries.ChartVariableCount = 0) then exit;   // no data

  CBEMAChartVisible(false);

  // default for harmonic chart only
  binSize := 5.0;

  channelCount := FChartTimeSeries.EventData.RawDataLabels.Count;  // number of channels to display

//  eventCount := GetActiveEventCount;
  eventCount := 0;
  if(FChartType = ctHarmonic) then
  begin
    harmSettings := FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsHRM;
  end else
  begin
    harmSettings := nil;
  end;

  // Allocate dynamic arrays used for charting

  AllocateChartArrays(channelCount, harmSettings, chartRec, eventRec, chartMarker,
                      eventNumber, eventColor, chartHeight, workArray, xArray,
                      table, vbase, binsize, doStatic);

  // Initialize chart axis scale

  InitChartMinMax(reset,xMin,xMax,yMin,yMax);

  // Initialize chart attributes

  InitChartAttributes(channelCount);

  // Fill data into chart series

  if((FChartType = ctLDP) or (FChartType = ctHarmonic) or
     (FChartTimeSeries.EventData.IndexTable.Count = 0)) then     // no events, display all data
  begin
    ChartLDPHarmonic(eventCount, channelCount,
                doStatic, vbase, binSize, harmSettings,
                chartMarker, chartRec,
                chartHeight, workArray,
                xArray, table);
  end else
  begin
  end;

  // Update display attributes

  FChartTimeSeries.InitChartTitles;
  FChartTimeSeries.InitChartActiveChannels;
  SetChartLineStyles;
  if(reset or ChartDirty) then
  begin
    ResetChartControls;   // update UI controls from present settings
  end;
  UpdateChartMinMax(reset,xMin,xMax,yMin,yMax);
  FChartDirty := false;     // the chart display is now up-to-date
  try
    if(UpdateEnabled) then
    begin
      ChartHMIDisplay.Update();  // refresh display
    end;
  except
    // swallow exception
  end;
end;

{************************************************************************
* NAME:   AllocateChartArrays
* FUNCTIONAL DESCRIPTION: allocate working arrays for charting
* INPUT PARAMETERS: channelCount : integer number of channels (chart series)
*                   harmSettings : TTimeSeriesInputSettingsHRM harmonic chart settings
* OUTPUT PARAMETERS:
*                   chartRec : TArrayOfInteger  sequential record number
*                   eventRec : TArrayOfInteger  actual record number
*                   chartMarker : TArrayOfInteger  seq. record for marker
*                   eventNumber : TArrayOfInteger  event # of marker
*                   eventColor : TArrayOfColor   color of marker
*                   chartHeight : TArrayOfSingle height of marker
*                   workArray : TArrayOfSingle  used to revise magnitude/angle arrays
*                   xArray : array of double indep. variable record number
*                   yArray : array of array of double dep. variable data
*                   vbase : single
*                   binsize : single
*                   doStatic : boolean true if not using dynamic data arrays
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.AllocateChartArrays(
              channelCount : integer;
              harmSettings : TTimeSeriesInputSettingsHRM;
              var chartRec, eventRec, chartMarker, eventNumber : TArrayOfInteger;
              var eventColor : TArrayOfColor;
              var chartHeight, workArray : TArrayOfDouble;
              var xArray : TArrayOfDouble;
              var yArray : TArrayOfArrayOfDouble;
              var vbase, binsize : double;
              var doStatic : boolean);
var
  dataRecord     : TTimeSeriesRecord;
  minbin : double;
  index, count : integer;
begin
  // Make sure old chart data is freed before reallocating arrays
  FreeCharts();
  // Event marker information (possible 4 markers per event)
  count := FChartTimeSeries.EventData.indexTable.Count;
  if(count > 0) then
  begin
    SetLength(chartRec,count);  // sequential record number
    SetLength(eventRec,count);  // actual record number
    SetLength(chartMarker,count*MARKERS_PER_EVENT); // seq. record for marker
    SetLength(eventNumber,count*MARKERS_PER_EVENT); // event # of marker
    SetLength(eventColor,count*MARKERS_PER_EVENT); // color of marker
    SetLength(chartHeight,count*MARKERS_PER_EVENT); // height of marker
  end;

  if(FChartType = ctHarmonic) then
  begin
    if(FChartTimeSeries.EventData.RawDataTable.Count > 1) then
    begin
      dataRecord := (FChartTimeSeries.EventData.RawDataTable[FChartTimeSeries.EventData.RawDataTable.Count-1] as TTimeSeriesRecord);
      binSize := dataRecord.RecordValue;   // floating representation of (frequency)
      dataRecord := (FChartTimeSeries.EventData.RawDataTable[0] as TTimeSeriesRecord);
      minBin := dataRecord.RecordValue;     // smallest frequence bin
      binSize := (binSize - minBin)/FChartTimeSeries.EventData.RawDataTable.Count;
    end else
    begin
      Assert(false,'Not enough records for Harmonic chart');
    end;

    if((FChartType = ctHarmonic) and
         ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
    begin
      SetLength(PolarChartSeries,FChartTimeSeries.ChartVariableCount div 2);
    end else
    begin
      SetLength(BarChartSeries,FChartTimeSeries.ChartVariableCount);
    end;
    SetLength(ChartSeries,0);

    if((FChartType = ctHarmonic) and
         ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
    begin
      SetLength(workArray,MA_ARRAY_SIZE);  // used to revise magnitude/angle arrays
    end else
    if((FChartType = ctHarmonic) and (FChartSubType = cstHRM_INT) and
       (harmSettings.HarmonicChartType <> htSpectral)) then
    begin
      SetLength(workArray,channelCount);  // used as spectral work array
    end;

    if(FChartTimeSeries.EventData.SettingTable.Count > 0) then
    begin
      vbase := (FChartTimeSeries.EventData.SettingTable[0] as TTimeSeriesSettingRecord).SettingValue;
    end else
    begin
      vbase := 0.0;
    end;
  end else
  begin
    SetLength(BarChartSeries,0);
    SetLength(ChartSeries,FChartTimeSeries.ChartVariableCount);
    vbase := 0.0;
  end;

  // Set TRUE if harmonic chart data is not filled into a dynamic array
  doStatic := ((FChartType = ctHarmonic) and ((FChartSubType = cstHRM_INT) and
                             (harmSettings.HarmonicChartType <> htSpectral)) or
               ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV)));

  // Build array if not doStatic because it is faster for large datasets
  if(not doStatic) then
  begin
    SetLength(xArray, FChartTimeSeries.EventData.RawDataTable.Count);
    SetLength(yArray, channelCount);

    for index := 0 to channelCount - 1 do
    begin
      SetLength(yArray[index],FChartTimeSeries.EventData.RawDataTable.Count);
    end;
  end;
end;

{************************************************************************
* NAME:   FillChartArrays
* FUNCTIONAL DESCRIPTION: fill TeeChart with data using arrays for speed
* INPUT PARAMETERS: channelCount : integer number of channels (chart series)
*                   recordCount : integer number of records
*                   xArray : array of double indep. variable record number
*                   yArray : array of array of double dep. variable data
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.FillChartArrays(channelCount, recordCount : integer;
              xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);
var
  index : integer;
begin
  // Send data to chart for each channel if using dynamic arrays for speed
  for index := 0 to channelCount - 1 do
  begin
    With FChartTimeSeries.ChartDisplay.Series[index].XValues do
    begin
      Value := TChartValues(xArray);
      Count := recordCount;
      Modified := true;
    end;
    With FChartTimeSeries.ChartDisplay.Series[index].YValues do
    begin
      Value := TChartValues(yArray[index]);
      Count := recordCount;
      Modified := true;
    end;
  end;
end;

{************************************************************************
* NAME:   ChartLDPHarmonic
* FUNCTIONAL DESCRIPTION: fill TeeChart arrays with data for LDP/Harmonic data
* INPUT PARAMETERS: eventCount : integer   active event count
*                   channelCount : integer number of channels (chart series)
*                   doStatic : boolean true if not using dynamic arrays
*                   vbase : single
*                   binSize : single   width of harmonic bin
*                   harmSettings : TTimeSeriesInputSettingsHRM
*                   chartMarker: TArrayOfInteger  seq. record for marker
*                   chartRec : TArrayOfInteger  sequential record number
*                   chartHeight : TArrayOfSingle   height of marker
*                   xArray : TArrayOfDouble      dynamic charting indep. var. arrays
*                   yArray : TArrayOfArrayOfDouble  dynamic charting dep var. arrays
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ChartLDPHarmonic(eventCount, channelCount : integer;
                doStatic : boolean; vbase, binSize : double;
                harmSettings : TTimeSeriesInputSettingsHRM;
                chartMarker, chartRec : TArrayOfInteger;
                chartHeight, workArray : TArrayOfDouble;
                xArray : TArrayOfDouble; yArray : TArrayOfArrayOfDouble);
const
  SKIP_RECORD = '(5)';
  TEST_RECORD = '(6)';
var
  index,j,rec : integer;
  myRecord : double;
  dataRecord        : TTimeSeriesRecord;
begin
  rec := 0;
  for index := 0 to FChartTimeSeries.EventData.RawDataTable.Count-1 do
  begin
    dataRecord := (FChartTimeSeries.EventData.RawDataTable[index] as TTimeSeriesRecord);
    if((dataRecord.Status = SKIP_RECORD) or (dataRecord.Status = TEST_RECORD)) then
    begin
      continue;  // don't display empty records
    end;
    if(FChartType = ctHarmonic) then
    begin
      myRecord := dataRecord.RecordValue;     // float representation
      if((FChartType = ctHarmonic) and
         ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
      begin
        ChartHarmonicMagAnglePolar(index,channelCount,myRecord,dataRecord);
      end else
      begin
        if(not SkipHarmonicRecord(myRecord)) then
        begin
          xArray[rec] := myRecord;
          for j := 0 to channelCount-1 do     // add every channel for this record to the chart
          begin
            yArray[j][rec] := dataRecord.Value[j];
          end;
          inc(rec);
        end;
      end;
    end else
    begin
      xArray[rec] := dataRecord.RecordNumber;
      for j := 0 to channelCount-1 do     // add every channel for this record to the chart
      begin
        yArray[j][rec] := dataRecord.Value[j];
      end;
      inc(rec);
    end;
  end;

  // Passing data to chart using arrays is faster for large data sets
  if(not doStatic) then
  begin
    FillChartArrays(channelCount, rec, xArray, yArray);
  end;

  ChartHarmonicMagAnglePolarLabels(channelCount);
end;

{************************************************************************
* NAME:   ChartHarmonicMagAnglePolar
* FUNCTIONAL DESCRIPTION: Save updated dependent values for Harmonic Mag/Angle
* INPUT PARAMETERS: index : integer record number
*                   channelCount : integer  number of dependent values
*                   myRecord : double independent value
*                   dataRecord : TTimeSeriesRecord   data record
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ChartHarmonicMagAnglePolar(index,channelCount : integer;
              myRecord : double;
              dataRecord : TTimeSeriesRecord);
var
  j : integer;
  mag, angle : array of Double;
begin
  if(SkipHarmonicRecord(myRecord)) then exit;
  SetLength(mag,channelCount div 2);
  SetLength(angle,channelCount div 2);

  for j := 0 to channelCount-1 do     // add every channel for this record to the chart
  begin
    if((j mod 2) = 0) then    // magnitude
    begin
      mag[j div 2] := dataRecord.Value[j];
      angle[j div 2] := dataRecord.Value[j+1];
    end;
  end;

  for j := 0 to channelCount div 2-1 do     // add every channel for this record to the chart
  begin
    FChartTimeSeries.ChartDisplay.Series[j].AddXY(angle[j],mag[j]);
  end;
end;

{************************************************************************
* NAME:   ChartHarmonicMagAnglePolar
* FUNCTIONAL DESCRIPTION: Save updated dependent values for Harmonic Mag/Angle
* INPUT PARAMETERS: index : integer record number
*                   channelCount : integer  number of dependent values
*                   myRecord : double independent value
*                   dataRecord : TTimeSeriesRecord   data record
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
function TfrmChartTimeSeries.SkipHarmonicRecord(myRecord : double) : boolean;
var
  skip : boolean;
  harmonicInput : TTimeSeriesInputSettingsHRM;
begin
  harmonicInput := FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsHRM;
  skip := false;
  if((harmonicInput.HarmonicChartDisplay = hdtOdd) and
     ((Round(myRecord) mod 2) = 0)) then
  begin
    skip := true;
  end else
  if((harmonicInput.HarmonicChartDisplay = hdtEven) and
     ((Round(myRecord) mod 2) = 1)) then
  begin
    skip := true;
  end else
  if((harmonicInput.HarmonicChartDisplay = hdtInterval) and
     ((Trunc(myRecord) < harmonicInput.HarmonicChartDisplayStart) or
     ((Trunc(myRecord) > (harmonicInput.HarmonicChartDisplayStart +
                          harmonicInput.HarmonicChartDisplayInterval - 1))))) then
  begin
    skip := true;
  end;
  result := skip;
end;

{************************************************************************
* NAME:   ChartHarmonicMagAnglePolarLabels
* FUNCTIONAL DESCRIPTION: Update labels on polar bar chart
* INPUT PARAMETERS:
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.ChartHarmonicMagAnglePolarLabels(channelCount : integer);
var
  series,index,rCount,activeRecs,i,j,itmp : integer;
  myRec : array of integer;
  angle, mag : array of Double;
  stmp : Double;
  myLabel : PWideChar;
  dataRecord        : TTimeSeriesRecord;
begin
  if((FChartType = ctHarmonic) and
     ((FChartSubType = cstHRM_MAI) or (FChartSubType = cstHRM_MAV))) then
  begin
    rCount := FChartTimeSeries.EventData.RawDataTable.Count;
    SetLength(myRec,rCount);
    SetLength(mag,rCount);
    SetLength(angle,rCount);

    for series := 0 to channelCount div 2 - 1 do
    begin
      activeRecs := 0;
      for index := 0 to rCount-1 do
      begin
        dataRecord := (FChartTimeSeries.EventData.RawDataTable[index] as TTimeSeriesRecord);
        myRec[activeRecs] := Trunc(dataRecord.RecordNumber);
        if(SkipHarmonicRecord(myRec[activeRecs])) then continue;

        mag[activeRecs] := dataRecord.Value[series*2];
        angle[activeRecs] := dataRecord.Value[series*2+1];
        Inc(activeRecs);
      end;

      // sort by harmonic

      for i := 0 to activeRecs - 2 do
      begin
        for j := i+1 to activeRecs - 1 do
        begin
          if(angle[i] > angle[j]) then
          begin
            itmp := myRec[i];
            myRec[i] := myRec[j];
            myRec[j] := itmp;
            stmp := mag[i];
            mag[i] := mag[j];
            mag[j] := stmp;
            stmp := angle[i];
            angle[i] := angle[j];
            angle[j] := stmp;
          end;
        end;
      end;

      // KWM Comment this block to avoid error during FreeAllSeries
      for i := 0 to activeRecs - 1 do
      begin
        myLabel := PWideChar(Format('%d=(%.1f,%.1f)',[myRec[i],mag[i],angle[i]]));
        PolarChartSeries[series].Marks[i].Text.SetText(myLabel);
//        PolarChartSeries[series].Marks[i].Text.Text := myLabel;
      end;
    end;
  end;
end;

{************************************************************************
* NAME:   GetCBEMA
* FUNCTIONAL DESCRIPTION: Get the CBEMA option from the input settings object
* INPUT PARAMETERS:
* OUTPUT PARAMETERS:
* RETURN: result : boolean  true if CBEMA active otherwise false
************************************************************************}
function TfrmChartTimeSeries.GetCBEMA : boolean;
begin
  if(FChartType = ctSSI) then
  begin
    result := (ChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI).ChartCBEMA;
  end else
  begin
    result := false;
  end;
end;

{************************************************************************
* NAME:   SetCBEMA
* FUNCTIONAL DESCRIPTION: Set the CBEMA option in the input settings object
* INPUT PARAMETERS: doCBEMA : boolean  true if CBEMA chart display is desired
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.SetCBEMA(doCBEMA : boolean);
begin
  if(FChartType = ctSSI) then
  begin
    (FChartTimeSeries.EventData.InputSettings as TTimeSeriesInputSettingsSSI).ChartCBEMA := doCBEMA;
  end;
end;

{************************************************************************
* NAME:  InitChartMinMax
* FUNCTIONAL DESCRIPTION:  initialize chart axis scaling
* INPUT PARAMETERS: reset : boolean  true if discarding current scale information
* OUTPUT PARAMETERS:
*                   xMin : Double  minimum value of x-axis scale
*                   xMax : Double  maximum value of x-axis scale
*                   yMin : Double  minimum value of y-axis scale
*                   yMax : Double  maximum value of y-axis scale
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartMinMax(reset : boolean; var xMin,xMax,yMin,yMax : Double);
begin
  if(reset) then   // throw away current scale information
  begin
    xMin := 0.0;
    xMax := 1.0;
    yMin := 0.0;
    yMax := 1.0;
  end else
  begin           // remember current axis limits
    xMin := FChartTimeSeries.ChartDisplay.BottomAxis.Minimum;
    xMax := FChartTimeSeries.ChartDisplay.BottomAxis.Maximum;
    yMin := FChartTimeSeries.ChartDisplay.LeftAxis.Minimum;
    yMax := FChartTimeSeries.ChartDisplay.LeftAxis.Maximum;
  end;
  FChartTimeSeries.ChartDisplay.BottomAxis.Automatic := false;
  FChartTimeSeries.ChartDisplay.LeftAxis.Automatic := false;
end;

{************************************************************************
* NAME:    UpdateChartMinMax
* FUNCTIONAL DESCRIPTION:   Update chart axis scaling after chart is finished
* INPUT PARAMETERS: reset : boolean  true if discarding current scale information
*                   xMin : Double  minimum value of x-axis scale
*                   xMax : Double  maximum value of x-axis scale
*                   yMin : Double  minimum value of y-axis scale
*                   yMax : Double  maximum value of y-axis scale
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.UpdateChartMinMax(reset : boolean; xMin,xMax,yMin,yMax : double);
const
  TOL = 1.0e-6;
begin
  if(reset) then
  begin
    FChartTimeSeries.AutoScaleAxes(false);   // use customized axis scaling
    if((FChartType = ctHarmonic) and (ChartScrollBarXAxis.Visible)) then
    begin
      ChartScrollBarXAxis.RecalcPosition;
    end;
  end else
  begin
    FChartTimeSeries.ChartDisplay.BottomAxis.Automatic := false;
    FChartTimeSeries.ChartDisplay.LeftAxis.Automatic := false;
    FChartTimeSeries.ChartDisplay.BottomAxis.Minimum := xMin;
    FChartTimeSeries.ChartDisplay.BottomAxis.Maximum := xMax;
    if((yMax - yMin) < TOL)  then    // Make sure there is some difference in the axis limits
    begin
      yMin := yMin - TOL;
      yMax := yMax + TOL;
    end;
    FChartTimeSeries.ChartDisplay.LeftAxis.Minimum := yMin;
    FChartTimeSeries.ChartDisplay.LeftAxis.Maximum := yMax;
  end;
end;

{************************************************************************
* NAME:   InitChartAttributes
* FUNCTIONAL DESCRIPTION:   Initialize attributes of the chart
* INPUT PARAMETERS: channelCount : integer   number of indep. variables
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.InitChartAttributes(channelCount : integer);
const
  CBEMA_TITLES : array [0..1] of string = ('ITI Sag','ITI Swell');
  EVENT_TITLE = 'Event';
var
  j : integer;
  textLabel : PWideChar;
begin
  textLabel := 'Channel';
  FChartTimeSeries.ChartDisplay.Title.Font.Color := clBlack;
  FChartTimeSeries.ChartDisplay.Title.Font.Name := 'Tahoma';
  FChartTimeSeries.ChartDisplay.Title.Font.Size := 12;
  FChartTimeSeries.ChartDisplay.BottomAxis.RoundFirstLabel := True;
  FChartTimeSeries.ChartDisplay.Legend.CheckBoxes := true;
  FChartTimeSeries.ChartDisplay.Legend.Title.Visible := true;
  FChartTimeSeries.ChartDisplay.Legend.Title.Text.Clear;
  FChartTimeSeries.ChartDisplay.Legend.Title.Text.SetText(textLabel);
  if((ChartType = ctHarmonic) and
     (ChartSubType = cstHRM_MAI) or (ChartSubType = cstHRM_MAV)) then
  begin
    FChartTimeSeries.ChartDisplay.BottomAxis.Title.Visible := false;
    FChartTimeSeries.ChartDisplay.LeftAxis.Title.Visible := false;
  end else
  begin
    FChartTimeSeries.ChartDisplay.BottomAxis.Title.Visible := true;
    FChartTimeSeries.ChartDisplay.LeftAxis.Title.Visible := true;
  end;


  if(FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime) then    // display date/time X axis labels
  begin
    With FChartTimeSeries.ChartDisplay.BottomAxis do
    begin
      LabelsAngle := 90;
    end;
  end else                               // display record number X axis labels
  begin
    With ChartTimeSeries.ChartDisplay.BottomAxis do
    begin
      LabelsAngle := 0;
      Automatic := true;
    end;
  end;

  // Allocate and initialize chart series

  for j := 0 to FChartTimeSeries.ChartVariableCount-1 do        // for each channel
  begin
    if(ChartType = ctHarmonic) then
    begin
      if((ChartSubType = cstHRM_MAI) or (ChartSubType = cstHRM_MAV)) then
      begin
        if(j < FChartTimeSeries.ChartVariableCount div 2) then  // combine mag/angle
        begin
          PolarChartSeries[j] := TPolarBarSeries.Create(self);    // create polar chart
        end;
      end else
      begin
        BarChartSeries[j] := TBarSeries.Create(self);    // create bar chart
      end;
    end else
    begin
      case FChartTimeSeries.EventData.InputSettings.ChartStyle of
        sctLine:
          begin
            ChartSeries[j] := TLineSeries.Create(self);    // create line chart
          end;
        sctArea:
          begin
            ChartSeries[j] := TAreaSeries.Create(self);   // create area chart
            TAreaSeries(ChartSeries[j]).Transparency := FChartTimeSeries.EventData.InputSettings.ChartTransparency;  // percent transparency
          end;
        else
        begin
          Assert(false,'Invalid chart style.');
        end;
      end;
    end;

    if(ChartType = ctHarmonic) then
    begin
      if((ChartSubType = cstHRM_MAI) or (ChartSubType = cstHRM_MAV)) then
      begin
        if(j < FChartTimeSeries.ChartVariableCount div 2) then  // combine mag/angle
        begin
          PolarChartSeries[j].Title := AnsiLeftStr(ChartTimeSeries.EventData.RawDataLabels[j*2],2);  // assign channel name to legend
          PolarChartSeries[j].Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
          PolarChartSeries[j].CircleLabels := true;
          PolarChartSeries[j].ClockWiseLabels := false;
          FChartTimeSeries.ChartDisplay.AddSeries(PolarChartSeries[j]);           // add the new channel series to the chart
        end;
      end else
      begin
        BarChartSeries[j].Title := ChartTimeSeries.EventData.RawDataLabels[j];  // assign channel name to legend
        BarChartSeries[j].Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
        FChartTimeSeries.ChartDisplay.AddSeries(BarChartSeries[j]);           // add the new channel series to the chart
      end;
    end else
    begin
      if(FChartTimeSeries.EventData.RawDataTable.Count > 50) then
      begin
        ChartSeries[j].AreaLinesPen.Visible := false;   // suppress grid lines
      end else
        ChartSeries[j].AreaLinesPen.Visible := true;

      if(j >= channelCount) then  // it is a CBEMA curve
      begin
        Assert(((j-channelCount) >= 0) and
                ((j-channelCount) < Length(CBEMA_TITLES)-1),
                'Array length mismatch');
        ChartSeries[j].Stairs := true;  // use step function style for CBEMA
        ChartSeries[j].Title := CBEMA_TITLES[j-channelCount];  // assign channel name to legend
      end else
      begin
        ChartSeries[j].Title := ChartTimeSeries.EventData.RawDataLabels[j];  // assign channel name to legend
      end;
      FChartTimeSeries.ChartDisplay.AddSeries(ChartSeries[j]);           // add the new channel series to the chart
    end;
  end;

  if((FChartType = ctSSI) and FChartTimeSeries.ChartHasEvents) then
  begin
    BarSeries := TBarSeries.Create(self);
    try
      BarSeries.Title := EVENT_TITLE;  // assign channel name to legend
      Barseries.Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
      Barseries.Marks.Clip := true;
      Barseries.Marks.Shadow.Clip := true;
      BarSeries.Marks.Shadow.Visible := false;   // hide shadow (doesn't clip)
      FChartTimeSeries.ChartDisplay.AddSeries(BarSeries);  // add the new series to the chart
    except
      // swallow exception
    end;
  end;

  if((ChartType = ctLDP) and ChartTimeSeries.EventsInInterval) then
  begin
    PointSeries := TPointSeries.Create(self);
    try
      PointSeries.Title := EVENT_TITLE;  // assign channel name to legend
      PointSeries.Marks.Visible := ChartTimeSeries.EventData.InputSettings.ChartLabels;
      PointSeries.Marks.Clip := true;
      PointSeries.Marks.Shadow.Clip := true;
      PointSeries.Marks.Arrowlength := 20;
      PointSeries.Marks.Shadow.Visible := false;   // hide shadow (doesn't clip)
      FChartTimeSeries.ChartDisplay.AddSeries(PointSeries);  // add the new series to the chart
    except
      // swallow exception
    end;
  end;
end;

{************************************************************************
* NAME:    FreeCharts
* FUNCTIONAL DESCRIPTION:   Free chart series
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.FreeCharts();
var
  count : integer;
begin
    try
      // There is a chronic problem related to freeing series in TeeChart.
      // This try/catch block is intended to minimize the impact of this problem.
      if(Assigned(ChartHMIDisplay)) then
      begin
        count := ChartHMIDisplay.SeriesCount();
        if(count > 0) then
        begin
          ChartHMIDisplay.FreeAllSeries();   // throw away previous chart data
        end;
      end;
      if(Assigned(ChartHMICBEMA)) then
      begin
        count := ChartHMICBEMA.SeriesCount();
        if(count > 0) then
        begin
          ChartHMICBEMA.FreeAllSeries();   // throw away previous chart data
        end;
      end;
    finally
      // nothing to do
    end;
end;

{!******************************************************************************
* MODULE NAME ...  SetDateTimeLabels
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  set chart date/time labels
*
* INPUT PARAMETERS:
*                    Sender : TObject
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.11.13  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SetDateTimeLabels();
const
  DATE_PT_LENGTH_LDP = 110;
  DATE_PT_LENGTH_SSI = 125;
  RECORD_PT_LENGTH = 15;
begin
  with FChartTimeSeries do
  begin

    if((EventData.ChartType = ctSSI) and     // Time/date not used for CBEMA
       TTimeSeriesInputSettingsSSI(EventData.InputSettings).ChartCBEMA) then
    begin
      exit;
    end;

    if(EventData.InputSettings.ChartAxisDateTime) then
    begin
      ChartDisplay.BottomAxis.LabelsAngle := 90;
      if(EventData.ChartType = ctSSI) then
      begin
        ChartDisplay.BottomAxis.LabelsSize := DATE_PT_LENGTH_SSI;
      end else
      begin
        ChartDisplay.BottomAxis.LabelsSize := DATE_PT_LENGTH_LDP;
      end;
    end else
    begin
      ChartDisplay.BottomAxis.LabelsAngle := 0;
      ChartDisplay.BottomAxis.LabelsSize := RECORD_PT_LENGTH;
    end;
    InitChartTitles();
    if(UpdateEnabled) then
    begin
      ChartDisplay.Update();
    end;
  end;
end;

{!******************************************************************************
* MODULE NAME ...  SetChartLineStyles
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Set chart line styles (color, type)
*
* INPUT PARAMETERS:
*                    None:
*
* OUTPUT PARAMETERS:
*                    None:
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.10.1  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.SetChartLineStyles();
var
  i , pen_width: integer;
  cPhase : THMIChannelPhase;
  cType : THMIChannelType;
begin

  if(FChartTimeSeries.ChartVariableCount = 0) then exit;   // no data
  if(FChartTimeSeries.EventData.ChartType = ctSSI) then
  begin
    if(TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA) then
    begin
      exit;
    end;
  end;

  if(Length(ChartSeries) < 1) then   // no data loaded into chart
  begin
    exit;
  end;

  // choose default pen widths.  Some channels will use different pen widths.
  if(FChartTimeSeries.ChartDisplay.View3D) then
  begin
    pen_width := 1;    // want thin lines for 3D
  end else
  begin
    pen_width := 3;    // want thick lines for 2D  (voltages and other channels)
  end;

  // update pen width for each channel
  for i := 0 to FChartTimeSeries.EventData.RawDataLabels.Count-1 do
  begin
    cPhase := FChartTimeSeries.GetChannelPhase(ChartTimeSeries.EventData.RawDataLabels[i]);
    cType := FChartTimeSeries.GetChannelType(ChartTimeSeries.EventData.RawDataLabels[i]);
    ChartSeries[i].LinePen.Width := pen_width;
    ChartSeries[i].SeriesColor := FChartTimeSeries.ChartLineColor(cPhase,i); // assign color to channel

    case FChartTimeSeries.EventData.InputSettings.ChartStyle of
      sctLine:
        begin
          if(cType = ctCurrent) then    // channel is current so use different pen width
          begin
            if(FChartTimeSeries.ChartDisplay.View3D) then
            begin
              ChartSeries[i].LinePen.Width := 1;    // want thin lines for 3D
            end else
            begin
              ChartSeries[i].LinePen.Width := 2;
            end;
          end;
          ChartSeries[i].AreaLinesPen.Style := FChartTimeSeries.ChartLineStyle(cType);
          ChartSeries[i].LinePen.Style := FChartTimeSeries.ChartLineStyle(cType);
        end;
      sctArea:
        begin
        end;
      else
        begin
          Assert(false,'Invalid chart style.');
        end;
    end;
  end;
end;

{************************************************************************
* NAME:    CBEMAChartVisible
* FUNCTIONAL DESCRIPTION: select type of chart to display for SSI
* INPUT PARAMETERS: doCBEMA : boolean  true if CBEMA chart is visible
* OUTPUT PARAMETERS: none
* RETURN: none     result : boolean   true if display should be refreshed
************************************************************************}
function TfrmChartTimeSeries.CBEMAChartVisible(doCBEMA : boolean) : boolean;
const
  SSI_TAB = 1;
var
  reset : boolean;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');

  FChartTimeSeries.EventData.InputSettings.ChartQuery := false;    // cancel cursor query
  SpeedButtonDateTime.Visible := not doCBEMA and (FChartType <> ctHarmonic);
  toolbar1.Realign;

  if(doCBEMA) then    // CBEMA
  begin
    TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA := true;
    ChartHMIDisplay.SendToBack;
    reset := (ChartHMICBEMA.SeriesCount = 0);
    FChartTimeSeries.ChartDisplay := ChartHMICBEMA;   // set active chart pointer
    FChartTimeSeries.CursorTool := ChartTool2;
  end else              // SSI/LDP
  begin
    TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA := false;
    ChartHMICBEMA.SendToBack;
    reset := false;  // no need to reset
    FChartTimeSeries.ChartDisplay := ChartHMIDisplay;   // set active chart pointer
    FChartTimeSeries.CursorTool := ChartTool1;
  end;
  FChartTimeSeries.CursorTool.Active := false;
  FChartTimeSeries.CursorTool.Snap := true;
  result := reset;
end;

procedure TfrmChartTimeSeries.ButtonImportClick(Sender: TObject);
begin
  if(comboBoxDateFormat.Text = 'MDY') then
  begin
    FFileCount := FFileCount + 1;
  end;

  if( FFileCount mod 3 = 1) then      // CHANGING ORDER CHANGES SYMPTOM
  begin
    FDefaultPath := 'PIV';
  end else
  if( FFileCount mod 3 = 2) then
  begin
    FDefaultPath := 'MAV';
  end else
  begin
    FDefaultPath := 'SSI';
  end;
  DisplayChartFile(FDefaultPath);
end;

{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayMouseMove
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  Event handler for mouse move events
*
* INPUT PARAMETERS:
*                    Sender : TObject
*                    Shift : TShiftState   set containing list of depressed control keys
*                    X,Y : Integer     screen coord. of mouse
*
* OUTPUT PARAMETERS:
*                    Handled: Boolean   true/false indicating whether event was handled
*
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
* 2008.12.18  Kent Myers        Add extra title for mouse cursor data feedback if necessary
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var
  deltaX, deltaY : integer;
  zoom2D : double;
begin
  Assert(FChartTimeSeries.EventData <> nil,'EventData is nil');
  Assert(FChartTimeSeries.EventData.InputSettings <> nil,'InputSettings is nil');

  deltaX := X - FChartTimeSeries.MouseX;
  deltaY := Y - FChartTimeSeries.MouseY;
  FChartTimeSeries.MouseX := X;
  FChartTimeSeries.MouseY := Y;
  if(((ssLeft in Shift) and
       FChartTimeSeries.ChartActive and
       (FChartTimeSeries.ChartRotate or
        FChartTimeSeries.ChartTranslate or
        FChartTimeSeries.ChartZoom or
        FChartTimeSeries.ChartThickness))) then
  begin
    with FChartTimeSeries.ChartDisplay.View3DOptions do
    begin
      if(FChartTimeSeries.EventData.InputSettings.Chart3D and
         FChartTimeSeries.ChartActive) then
      begin
        if(FChartTimeSeries.ChartRotate) then
        begin
          Orthogonal := false;
          Rotation := Rotation + deltaX;
          Elevation := Elevation - deltaY;
        end;
        if(FChartTimeSeries.ChartTranslate) then
        begin
          HorizOffset := HorizOffset + deltaX;
          VertOffset := VertOffset + deltaY;
        end;
        if(FChartTimeSeries.ChartZoom) then
        begin
          Zoom := Zoom + deltaX;
          Zoom := Zoom - deltaY;
        end;
        if(FChartTimeSeries.ChartThickness) then
        begin
          if(((FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX) < 101) and
             ((FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX) > 0)) then
          begin
             FChartTimeSeries.ChartDisplay.Chart3DPercent :=
                 FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX;
          end else
          begin
            if((FChartTimeSeries.ChartDisplay.Chart3DPercent + deltaX) >= 101) then
            begin
              FChartTimeSeries.ChartDisplay.Chart3DPercent := 100;
            end else
            begin
              FChartTimeSeries.ChartDisplay.Chart3DPercent := 1;
            end;
          end;
        end;
      end;
    end;
  end else
  begin
    if((ssRight in Shift) and ((ssCtrl in Shift) or (ssAlt in Shift) or (ssShift in Shift))) then
    begin
      with FChartTimeSeries.ChartDisplay.View3DOptions do
      begin
        if(ssShift in Shift) then     // do 3D rotation
        begin
          Elevation := Elevation - deltaY;
          Rotation := Rotation + deltaX;
        end else
        if(ssCtrl in Shift) then      // do 2D zoom
        begin
          zoom2D := 100.0 - deltaY;
          FChartTimeSeries.ChartDisplay.ZoomPercent(zoom2D);
        end else
        if(ssAlt in Shift) then       // do 3D zoom
        begin
          FChartTimeSeries.ChartDisplay.View3DOptions.Zoom :=
              FChartTimeSeries.ChartDisplay.View3DOptions.Zoom - deltaY;
          FChartTimeSeries.ChartDisplay.View3DOptions.Zoom :=
              FChartTimeSeries.ChartDisplay.View3DOptions.Zoom + deltaX;
        end;
      end;
    end else
    begin   // get X,Y values from curve
      DisplayMouseQueryFeedback;
    end;
  end;
  FChartTimeSeries.EventData.InputSettings.Rotation := FChartTimeSeries.ChartDisplay.View3DOptions.Rotation;
  FChartTimeSeries.EventData.InputSettings.Zoom := FChartTimeSeries.ChartDisplay.View3DOptions.Zoom;
  FChartTimeSeries.EventData.InputSettings.HorizOffset := FChartTimeSeries.ChartDisplay.View3DOptions.HorizOffset;
  FChartTimeSeries.EventData.InputSettings.VertOffset := FChartTimeSeries.ChartDisplay.View3DOptions.VertOffset;
  FChartTimeSeries.EventData.InputSettings.Elevation := FChartTimeSeries.ChartDisplay.View3DOptions.Elevation;
  FChartTimeSeries.EventData.InputSettings.Percent3D := FChartTimeSeries.ChartDisplay.Chart3DPercent;
end;

{************************************************************************
* NAME:   DisplayMouseQueryFeedback
* FUNCTIONAL DESCRIPTION:  Display feedback for chart object at mouse cursor location
* INPUT PARAMETERS: none
* OUTPUT PARAMETERS: none
* RETURN: none
************************************************************************}
procedure TfrmChartTimeSeries.DisplayMouseQueryFeedback;
var
  P:TPoint;
  found, isCBEMA : boolean;
  myX, myY : double;
  i, j, count, nActiveRecords, match, index, lastRecord : integer;
  indexRecord : TTimeSeriesIndexRecord;
  dataRecord : TTimeSeriesRecord;
  myLabel, myDT : string;
  myFormat : TFormatSettings;
begin
  FChartTimeSeries.EventData.InitDateTimeFormat(ftInternational,myFormat);

  isCBEMA := false;
  case FChartType of
    ctUndefined, ctLDP, ctHarmonic:
    begin
    end;
    ctSSI:
    begin
      if(TTimeSeriesInputSettingsSSI(FChartTimeSeries.EventData.InputSettings).ChartCBEMA) then
      begin
        isCBEMA := true;
      end;
    end;
    else
    begin
      Assert(false,'Invalid chart type');
    end;
  end;

  {NOTE: A semi-colon separator is used (instead of a comma) when feeding back
   a list of data values in order to avoid confusion when the native language
   uses a comma instead of a decimal point}

  if(FChartTimeSeries.EventData.InputSettings.ChartQuery) then
  begin
    if(FChartTimeSeries.ChartDisplay.Title.Text.Count <= 1) then
    begin
      FChartTimeSeries.ChartDisplay.Title.Text.Add(' ');   // add title for feedback
    end;
    P:=FChartTimeSeries.ChartDisplay.GetCursorPos;

    // Compare mouse location to each point in each series

    for i := 0 to FChartTimeSeries.ChartDisplay.SeriesCount - 1 do
    begin
      match := FChartTimeSeries.ChartDisplay.Series[i].Clicked(P.X,P.Y);
      if match<>-1 then     // match is matching record in series
      begin
        FChartTimeSeries.CursorTool.Series := FChartTimeSeries.ChartDisplay.Series[i];
        if((not isCBEMA) and
            not (FChartTimeSeries.ChartDisplay.Series[i] is TPolarBarSeries) and
            not (FChartTimeSeries.ChartDisplay.Series[i] is TBarSeries)) then
        begin
          match := FChartTimeSeries.CursorTool.SnapToPoint();   // not a point series, must identify data point
        end;
        myX := FChartTimeSeries.ChartDisplay.Series[i].XValue[match];
        myY := FChartTimeSeries.ChartDisplay.Series[i].YValue[match];

        if((FChartType = ctLDP) and (ChartTimeSeries.ChartDisplay.Series[i] is TPointSeries)) then
        begin
          index := FChartTimeSeries.GetEventRecordFromRecord(myX);
          if(index >= 0) then
          begin
            indexRecord := (FChartTimeSeries.EventData.IndexTable[index] as TTimeSeriesIndexRecord);
            myDT := DateToStr(indexRecord.DateTime,myFormat) + ' ' + FormatDateTime(TIME_WITH_MS,indexRecord.DateTime);
            myLabel := Format('Event %s (%s) = %s',
                        [indexRecord.EventName,myDT,indexRecord.Description]);
          end;
        end else
        if(not isCBEMA and FChartTimeSeries.EventData.InputSettings.ChartAxisDateTime) then    // display date/time X axis labels
        begin      // Indep. Var. is Date/Time
          // Determine which record this data point refers to
          found := false;
          if(ChartType = ctSSI) then // update match based on event records
          begin
            nActiveRecords := 0;
            lastRecord := 0;   // upper limit of last active parent record
            for j := 0 to FChartTimeSeries.EventData.IndexTable.Count - 1 do
            begin
              indexRecord := (FChartTimeSeries.EventData.IndexTable[j] as TTimeSeriesIndexRecord);  // get event summary record
              if(indexRecord.LastRecord <= lastRecord) then continue;  // skip sub-event
              lastRecord := indexRecord.LastRecord; // remember last record of active parent event
              count := indexRecord.LastRecord - indexRecord.FirstRecord + 1;  // number of records in this event
              if(not found and (match > nActiveRecords) and
                (match <= (nActiveRecords + count))) then  // record is in this interval
              begin
                found := true;      // record is in this event interval
                match := indexRecord.FirstRecord + match - nActiveRecords -1;
                break;
              end;
              Inc(nActiveRecords,count);  // number of active records
            end;
          end else
          begin
            found := true;  // LDP just use match already found
          end;

          if(found) then
          begin
            dataRecord := (FChartTimeSeries.EventData.RawDataTable[match] as TTimeSeriesRecord);
            if(FChartType = ctSSI) then
            begin
              myDT := DateToStr(dataRecord.DateTime,myFormat) + ' ' + FormatDateTime(TIME_WITH_MS,dataRecord.DateTime);
            end else
            begin
              myDT := DateToStr(dataRecord.DateTime,myFormat) + ' ' + FormatDateTime(TIME_WITH_SEC,dataRecord.DateTime);
            end;
            myLabel := Format('%s=(%s; %.2f)',[FChartTimeSeries.ChartDisplay.Series[i].Title,myDT,myY]);
          end;
        end else
        begin        // Indep. var. is record number or duration (if CBEMA)
          if(ChartType = ctLDP) then
          begin
            myLabel := Format('%s=(%.0f; %.2f)',[FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
          end else
          if(ChartType = ctHarmonic) then
          begin
            if((ChartSubType = cstHRM_MAI) or (ChartSubType = cstHRM_MAV)) then
            begin
              myLabel := Format('%s - %s',
                 [FChartTimeSeries.ChartDisplay.Series[i].Title,
                  Trim(FChartTimeSeries.ChartDisplay.Series[i].Marks[match].Text.GetText)]);
            end else
            begin
              myLabel := Format('%s=(%.0f; %.3f)',[FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
            end;
          end else
          begin
            if(isCBEMA) then     // display CBEMA event marker info
            begin
              if(i = 0) then    // first series is the event marker
              begin
                myLabel := Format('%s %s=(Duration=%.3f; Magnitude=%.1f)',
                [FChartTimeSeries.ChartDisplay.Series[i].Title,
                ChartTimeSeries.ChartDisplay.Series[i].Marks[match].Text.GetText,myX,myY]);
//                ChartTimeSeries.ChartDisplay.Series[i].Marks[match].Text[0],myX,myY]);
              end else           // display sag/swell limit info
              begin
                myLabel := Format('%s =(Duration=%.3f; Magnitude=%.1f)',
                [FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
              end;
            end else             // display SSI channel series info
            begin
              myLabel := Format('%s=(%.0f; %.2f)',   // Rec #, Magnitude
              [FChartTimeSeries.ChartDisplay.Series[i].Title,myX,myY]);
            end;
          end;
        end;

        FChartTimeSeries.ChartDisplay.Title.Text[FChartTimeSeries.ChartDisplay.Title.Text.Count-1] := myLabel;
      end;
    end;
  end;
end;

{!******************************************************************************
* MODULE NAME ...  ChartHMIDisplayGetNextAxisLabel
* SYSTEM NAME ...  AcSELerator
* AUTHOR ........  Kent Myers
*
* FUNCTIONAL DESCRIPTION:  called just before displaying axis labels
*
* INPUT PARAMETERS:
*                    Sender : TChartAxis
*                    LabelIndex : Integer  index of current label
*                    LabelValue : double   value of current label
*                    Stop  : boolean       true/false to stop callback
*
* OUTPUT PARAMETERS:
*                    None:
* RETURN VALUE:
*                    None:
*
* ERRORS:
*
* APPLICATION NOTES & EXAMPLE:
*  NOTE: This callback is necessary in order to obtain the increment between
*  consecutive labels.  If CalcIncrement() is called in the
*  ChartHMIDisplayGetAxisLabel callback, a recursion occurs during a zoom.
*
* REVISION HISTORY:
* REVISION    ORIGINATOR/       REVISION
* --DATE--    IMPLEMENTER       DESCRIPTION
* 2008.09.24  Kent Myers        Original implementation
*******************************************************************************}
procedure TfrmChartTimeSeries.ChartHMIDisplayGetNextAxisLabel(Sender: TChartAxis;
  LabelIndex: Integer; var LabelValue: Double; var Stop: Boolean);
begin
  if(Sender = ChartHMIDisplay.BottomAxis) then
  begin
    FChartTimeSeries.ChartXDelta := Sender.CalcIncrement;
  end;
end;

initialization
  RegisterClass(TfrmChartTimeSeries);
end.

